#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ncurses.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#define	MAIN_Y			4
#define	MAIN_X			26

#define	MAXFILENAME		40
#define	DEFAULT_HEX_RECORD_LEN	0x20

#define	LOAD_BIN		0
#define	LOAD_HEX		1
#define	SAVE_BIN		2
#define	SAVE_HEX		3
#define	READ			4
#define	WRITE			5
#define	CLEAR			6
#define	EDIT			7
#define	FILL			8
#define	RUN			9
#define	QUIT			10

#define TOP             	0
#define MIDDLE          	1
#define BOTTOM          	2


#define	WHITE_BLUE		1
#define	BLACK_WHITE		2
#define	WHITE_BLACK		3
#define	WHITE_RED		4
#define	WHITE_GREEN		5


#define	MEMORY_SIZE		0xFFFF	// 65,535


unsigned char	mbuffer[MEMORY_SIZE];

/*
COLOR_BLACK    COLOR_RED        COLOR_GREEN    COLOR_YELLOW
COLOR_BLUE     COLOR_MAGENTA    COLOR_CYAN     COLOR_WHITE

A_NORMAL       A_STANDOUT       A_UNDERLINE    A_REVERSE
A_BLINK        A_DIM A_BOLD     A_ALTCHARSET   A_INVIS
*/



int	ElfMon();

int	loadBinary(char *filename, int *buf_start, int *buf_end);
int	loadHex(char *filename, int *buf_start, int *buf_end);
int	saveBinary(char *filename, int *buf_start, int *buf_end);
int	saveHex(char *filename, int *buf_start, int *buf_end);

int	loadBinaryFile(char *filename, int *buf_start, int *buf_end);
int	saveBinaryFile(char *filename, int *buf_start, int *buf_end);

int	loadHexFile(char *filename, int *buf_start, int *buf_end);
int	decodeHex(char *buffer, int *b_start, int *buf_end);
int	getByte(char *buffer, unsigned char *sum);
int	hexVal(char hexch);

int	saveHexFile(char *filename, int *buf_start, int *buf_end, int defLen);

int	displayMessage(char *message, int mode, int attr);

int	editFileOptions(char *filename, int *buf_start, int *buf_end, int *fill, int mode, int attr);
int	endMove();

int	lined(char *buffer, int y, int x, int max, int attr, int alpha_num, int disable);
int	insert(char ch, char *buffer, int pos, int len);
int	backspace(char *buffer, int pos, int len);

int	setHighlite(int y, int x, int len, int attr, int effect);
int	displayFile(int y, int x, int attr, char *filename);
int	displayBufEnd(int y, int x, int attr, int buf_end);
int	menu_1(int cdypos, int cdxpos, int ylen, int xlen, int attr);
int	mborder(int cdypos, int cdxpos, int ylen, int xlen, int attr);
int	eborder(int cdypos, int cdxpos, int ylen, int xlen, int attr, int part1, int part2);

int	fillMbuffer(unsigned char *buffer, int start, int len, unsigned char fill);

int	fllBufferArea(int attr);
int	clearBuffer(char *filename, int *buf_start, int *buf_end, int attr);

int	hexed(unsigned char *image, int *length);
int	getaddress(int max, int y, int x);
int	cursor_right(int *nibble, int *y, int *x, int *offset);
int	printimage(unsigned char *image, int offset);
char	*hexstr(int value, int digits);
void	help();


int	main()
	{
	fillMbuffer(mbuffer, 0, MEMORY_SIZE, 0x00);

        initscr();
        keypad(stdscr, TRUE);
        start_color();
        noecho();
        raw();
        clear();
	curs_set(0); // 0,1,2
        refresh();

        init_pair(WHITE_BLUE, COLOR_WHITE, COLOR_BLUE);
        init_pair(BLACK_WHITE, COLOR_BLACK, COLOR_WHITE);
        init_pair(WHITE_BLACK, COLOR_WHITE, COLOR_BLACK);
        init_pair(WHITE_RED, COLOR_WHITE, COLOR_RED);
        init_pair(WHITE_GREEN, COLOR_WHITE, COLOR_GREEN);

	ElfMon();

        clear();
        refresh();
        endwin();
	return(0);
	}


int	ElfMon()
	{
	int	loop = 1, pos = 0;
	int	buf_start = 0, buf_end = 0;
	char	filename[64];


	filename[0] = 0;

	menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
	setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
	displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
	displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);

	while(loop)
		{
		switch(getch())
			{
			case 10:
				switch(pos)
					{
					case LOAD_BIN:
						loadBinary(filename, &buf_start, &buf_end);
						menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
						setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
						displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
						displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);
						break;
					case LOAD_HEX:
						loadHex(filename, &buf_start, &buf_end);
						menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
						setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
						displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
						displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);
						break;
					case SAVE_BIN:
						saveBinary(filename, &buf_start, &buf_end);
						menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
						setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
						displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
						displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);
						break;
					case SAVE_HEX:
						saveHex(filename, &buf_start, &buf_end);
						menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
						setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
						displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
						displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);
						break;
					case READ:
						break;
					case WRITE:
						break;
					case CLEAR:
						clearBuffer(filename, &buf_start, &buf_end, WHITE_RED);
						menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
						setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
						displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
						displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);
						break;
					case EDIT:
						hexed(mbuffer, &buf_end);
						menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
						setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
						displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
						displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);
						break;
					case FILL:
						fllBufferArea(BLACK_WHITE);
						menu_1(MAIN_Y,MAIN_X,14,27,WHITE_BLUE);
						setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
						displayFile(MAIN_Y + 14, MAIN_X + 2, WHITE_BLUE, filename);
						displayBufEnd(MAIN_Y + 14, MAIN_X + 18, WHITE_BLUE, buf_end);
						break;
					case RUN:
						break;
					case QUIT:
						loop = 0;
						break;
					default:
						break;
					}
				break;

			case 'c':
			case 'C':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);
				pos = CLEAR;
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 'e':
			case 'E':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);
				pos = EDIT;
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 'f':
			case 'F':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);
				pos = FILL;
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 'l':
			case 'L':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);

				if(pos == LOAD_BIN)
					pos = LOAD_HEX;
				else if (pos == LOAD_HEX)
					pos = LOAD_BIN;
				else
					pos = LOAD_BIN;

				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 'q':
			case 'Q':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);
				pos = QUIT;
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 'r':
			case 'R':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);

				if(pos == READ)
					pos = RUN;
				else if (pos == RUN)
					pos = READ;
				else
					pos = READ;

				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 's':
			case 'S':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);

				if(pos == SAVE_BIN)
					pos = SAVE_HEX;
				else if (pos == SAVE_HEX)
					pos = SAVE_BIN;
				else
					pos = SAVE_BIN;

				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 'w':
			case 'W':
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);
				pos = WRITE;
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			case 258:               /* DOWN */
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);

				if(pos < 10)
					++pos;
				else
					pos = 0;

				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
                                break;
			case 259:               /* UP */
				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_NORMAL);

				if(pos > 0)
					--pos;
				else
					pos = 10;

				setHighlite(MAIN_Y + 2 + pos, MAIN_X + 4, 20, WHITE_BLUE, A_REVERSE);
				break;
			default:
				break;
			}
		}


	return(0);
	}


// WORKING
int	loadBinary(char *filename, int *buf_start, int *buf_end)
	{
	int	fill = 0;
	char	message[32];


	if(editFileOptions(filename, buf_start, buf_end, &fill, LOAD_BIN, BLACK_WHITE))
		{
		if(fill > 0)
			fillMbuffer(mbuffer, 0, MEMORY_SIZE, fill);

		switch(loadBinaryFile(filename, buf_start, buf_end))
			{
			case 0:
				if(fill > 0)
					sprintf(message, "File Loaded with fill [%02X]", fill);
				else
					sprintf(message, "File Loaded - no fill");

				displayMessage(message , LOAD_BIN, WHITE_GREEN);
				break;
			case 1:
				displayMessage("Unable to open file" , LOAD_BIN, WHITE_RED);
				break;
			default:
				break;
			}
		}

	return(0);
	}


int	loadHex(char *filename, int *buf_start, int *buf_end)
	{
	int	fill = 0;
	char	message[32];


	if(editFileOptions(filename, buf_start, buf_end, &fill, LOAD_HEX, BLACK_WHITE))
		{
		if(fill > 0)
			fillMbuffer(mbuffer, 0, MEMORY_SIZE, fill);

		switch(loadHexFile(filename, buf_start, buf_end))
			{
			case 0:
				if(fill > 0)
					sprintf(message, "File Loaded with fill [%02X]", fill);
				else
					sprintf(message, "File Loaded - no fill");

				displayMessage(message , LOAD_HEX, WHITE_GREEN);
				break;
			case 1:
				displayMessage("Unable to open file" , LOAD_HEX, WHITE_RED);
				break;
			case 2:		// Bad record Format
				displayMessage("Bad record Format" , LOAD_HEX, WHITE_RED);
				break;
			case 3:		// Byte Count Error
				displayMessage("Byte Count Error" , LOAD_HEX, WHITE_RED);
				break;
			case 4:		// Record Address Error
				displayMessage("Record Address Error" , LOAD_HEX, WHITE_RED);
				break;
			case 5:		// Type Error
				displayMessage("Record Type Error" , LOAD_HEX, WHITE_RED);
				break;
			case 6:		// Unsupported type
				displayMessage("Unsupported record type" , LOAD_HEX, WHITE_RED);
				break;
			case 7:		// Data error
				displayMessage("Data error" , LOAD_HEX, WHITE_RED);
				break;
			case 8:		// missing checksum
				displayMessage("missing checksum" , LOAD_HEX, WHITE_RED);
				break;
			case 9:		// Checksum missmatch
				displayMessage("Checksum missmatch" , LOAD_HEX, WHITE_RED);
				break;
			case 10:	// Missing EOF record
				displayMessage("Missing EOF record" , LOAD_HEX, WHITE_RED);
				break;
			default:
				break;
			}

		}

	return(0);
	}


int	saveBinary(char *filename, int *buf_start, int *buf_end)
	{
	int	fill = 0;


	if(editFileOptions(filename, buf_start, buf_end, &fill, SAVE_BIN, BLACK_WHITE))
		{
		switch(saveBinaryFile(filename, buf_start, buf_end))
			{
			case 0:
				displayMessage("File Loaded" , SAVE_BIN, WHITE_GREEN);
				break;
			case 1:
				displayMessage("Unable to open file" , SAVE_BIN, WHITE_RED);
				break;
			default:
				break;
			}
		}

	return(0);
	}


int	saveHex(char *filename, int *buf_start, int *buf_end)
	{
	int	fill = 0;


	if(editFileOptions(filename, buf_start, buf_end, &fill, SAVE_HEX, BLACK_WHITE))
		{
		switch(saveHexFile(filename, buf_start, buf_end, DEFAULT_HEX_RECORD_LEN))
			{
			case 0:
				displayMessage("File Loaded" , SAVE_HEX, WHITE_GREEN);
				break;
			case 1:
				displayMessage("Unable to open file" , SAVE_HEX, WHITE_RED);
				break;
			default:
				break;
			}
		}

	return(0);
	}


int	loadBinaryFile(char *filename, int *buf_start, int *buf_end)
	{
	int		fptr, len;
	unsigned char	*mptr;


	mptr = mbuffer + *buf_start;

	if(-1 == (fptr = open(filename, O_RDONLY)))
		return(1);

	len = read(fptr, mptr, MEMORY_SIZE);

	close(fptr);

	*buf_end = *buf_start + len -1;

	return(0);
	}


int	saveBinaryFile(char *filename, int *buf_start, int *buf_end)
	{
	int		fptr, len;
	unsigned char	*mptr;


	len = (*buf_end - *buf_start) +1;

	mptr = mbuffer + *buf_start;

	if(-1 == (fptr = open(filename, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)))
		return(1);

	len = write(fptr, mptr, len);

	close(fptr);

	return(0);
	}


// WORKHEX
int	loadHexFile(char *filename, int *buf_start, int *buf_end)
	{
	int	status, abort = 0, b_start, line = 0;
	char	buffer[1024];
	FILE	*fptr;


	if(NULL != (fptr = fopen(filename, "r")))
		{
		while(!abort && fgets(buffer, 1023, fptr))
			{
			status = decodeHex(buffer, &b_start, buf_end);

			switch(status)
				{
				case 0:		// Data record
					if(!line)
						{
						++line;
						*buf_start = b_start;
						}
					break;
				case 1:		// EOF Record
					abort = 1;
					break;
				case 2:		// Bad record Format
					abort = 1;
					break;
				case 3:		// Byte Count Error
					abort = 1;
					break;
				case 4:		// Record Address Error
					abort = 1;
					break;
				case 5:		// Type Error
					abort = 1;
					break;
				case 6:		// Unsupported type
					abort = 1;
					break;
				case 7:		// Data error
					abort = 1;
					break;
				case 8:		// missing checksum
					abort = 1;
					break;
				case 9:		// Checksum missmatch
					abort = 1;
					break;
				default:
					break;
				}

			}
		}
	else
		return(1);  // Unable to open file

	fclose(fptr);

	if(!abort && !status)
		status = 10;	// Missing EOF record
	else if(abort && status == 1)
		status = 0;
	
	return(status);
	}


int	decodeHex(char *buffer, int *b_start, int *buf_end)
	{
	int		i;
	int		count, address, type, data, checksum;
	unsigned char	*mptr, sum = 0, csum;


	if(*buffer != ':')
		return(2);

	++buffer;
	
	if((count = getByte(buffer, &sum)) == -1)
		return(3);

	buffer += 2;

	if((address = getByte(buffer, &sum)) == -1)
		return(4);

	buffer += 2;

	address <<= 8;

	if((address |= getByte(buffer, &sum)) == -1)
		return(4);

	buffer += 2;

	*b_start = address;


	if((type = getByte(buffer, &sum)) == -1)
		return(5);

	if(type > 1)
		return(6);

	buffer += 2;

	mptr = mbuffer + address;

	for(i = 0 ; i < count ; ++i)
		{
		if((data = getByte(buffer, &sum)) == -1)
			return(7);

		*mptr++ = data;

		buffer += 2;

		*buf_end = address++;
		}

	csum = ~sum +1;

	if((checksum = getByte(buffer, &sum)) == -1)
		return(8);

	if(csum != checksum)
		return(9);

	return(type);
	}


int	getByte(char *buffer, unsigned char *sum)
	{
	int	data, tmp;


	if(*buffer)
		{
		if((tmp = hexVal(*buffer++)) == -1)
			return(-1);

		data = tmp << 4;
		}
	else
		return(-1);

	if(*buffer)
		{
		if((tmp = hexVal(*buffer)) == -1)
			return(-1);

		data += tmp;
		}
	else
		return(-1);

	*sum += data;

	return(data);
	}


int	hexVal(char hexch)
	{
	if(hexch >= '0' && hexch <= '9')
		return(hexch - '0');
	else if(hexch >= 'A' && hexch <= 'F')
		return(hexch - 55);
	else if(hexch >= 'a' && hexch <= 'f')
		return(hexch - 87);

	return(-1);
	}


int	saveHexFile(char *filename, int *buf_start, int *buf_end, int defLen)
	{
	int		i, total = (*buf_end - *buf_start) + 1;
	int		count, address, type;
	unsigned char	*mptr, sum, checksum;
	FILE		*fptr;


	address = *buf_start;

	mptr = mbuffer + address;

	if(NULL != (fptr = fopen(filename, "w")))
		{
		while(total)
			{
			if(total > defLen)
				count = defLen;
			else
				count = total;

			type = 0x00;

			sum = 0x00;
			sum += count;	
			sum += (address >> 8) & 0xff;	
			sum += address & 0xFF;	
			sum += type;	

			fprintf(fptr, ":%02X%04X%02X", count, address, type);

			for(i = 0 ; i < count ; ++i)
				{
				sum += *mptr;
				fprintf(fptr, "%02X", *mptr++);
				}

			checksum = ~(sum -1);
			fprintf(fptr, "%02X\n", checksum);

			total -= count;
			address += count;
			}

		count = 0x00;
		type = 0x01;

		sum = 0;
		sum += count;	
		sum += (address >> 8) & 0xFF;	
		sum += address & 0xFF;	
		sum += type;	

		checksum = ~(sum -1);

		fprintf(fptr, ":%02X%04X%02X%02X\n", count, address, type, checksum);			

		fclose(fptr);
		}
	else
		return(1);

	return(0);
	}


int	displayMessage(char *message, int mode, int attr)
	{
	mborder(11 + mode, 20, 4, 40, attr);	// Edit Window

	attron(COLOR_PAIR(attr));
	mvprintw(13 + mode, 40 - strlen(message) / 2, message);
	attroff(COLOR_PAIR(attr));

	refresh();

	setHighlite(13 + mode, 21, 39, attr, A_BOLD);

	getchar();

	return(0);
	}


int	editFileOptions(char *filename, int *buf_start, int *buf_end, int *fill, int mode, int attr)
	{
	int	abort = 0, edit = 1, pos = 0, status = 0;
	char	buffer[16];


	mborder(MAIN_Y + 3 + mode, MAIN_X - 13, 12, 54, attr);	// Edit Window

	mborder(MAIN_Y + 4 + mode, MAIN_X - 6, 2, 42, attr);	// Filename 

	mborder(MAIN_Y + 9 + mode, MAIN_X - 10, 2, 9, attr);	// Bstart
	mborder(MAIN_Y + 9 + mode, MAIN_X + 10, 2, 9, attr);	// Bend

	if(mode == LOAD_BIN || mode == LOAD_HEX)
		mborder(MAIN_Y + 9 + mode, MAIN_X + 30, 2, 9, attr);	// Fill

	mborder(MAIN_Y + 12 + mode, MAIN_X + 0, 2, 9, attr);	// Cancel
	mborder(MAIN_Y + 12 + mode, MAIN_X + 20, 2, 9, attr);	// Load/Save

	attron(COLOR_PAIR(attr));

	mvprintw(MAIN_Y + 5 + mode, MAIN_X - 11,"FILE");

	mvprintw(MAIN_Y + 8 + mode, MAIN_X - 11,"BUFFER START");
	mvprintw(MAIN_Y + 8 + mode, MAIN_X + 10,"BUFFER END");

	if(mode == LOAD_BIN || mode == LOAD_HEX)
		mvprintw(MAIN_Y + 8 + mode, MAIN_X + 29,"BUFFER FILL");

	mvprintw(MAIN_Y + 13 + mode, MAIN_X + 2,"CANCEL");

	switch(mode)
		{
		case LOAD_BIN:
		case LOAD_HEX:
			mvprintw(MAIN_Y + 13 + mode, MAIN_X + 23,"LOAD");
			break;
		case SAVE_BIN:
		case SAVE_HEX:
			mvprintw(MAIN_Y + 13 + mode, MAIN_X + 23,"SAVE");
			break;
		}



	// Fill in the Data

	mvprintw(MAIN_Y + 5 + mode, MAIN_X - 5,filename);

	sprintf(buffer, "0x%04X", *buf_start);
	mvprintw(MAIN_Y + 10 + mode, MAIN_X - 8,buffer);

	sprintf(buffer, "0x%04X", *buf_end);
	mvprintw(MAIN_Y + 10 + mode, MAIN_X + 12,buffer);

	if(mode == LOAD_BIN || mode == LOAD_HEX)
		{
		sprintf(buffer, "0x%02X", *fill);
		mvprintw(MAIN_Y + 10 + mode, MAIN_X + 33,buffer);
		}


	while(edit)
		{
		switch(pos)
			{
			case 0:	// Filename
				if(lined(filename, MAIN_Y + 5 + mode, MAIN_X - 5, MAXFILENAME, attr, 1, 0))
					{
					abort = 1;
					edit = 0;
					}
				++pos;
				break;
			case 1:	// Buffer Start
				sprintf(buffer, "%04X", *buf_start);
				if(lined(buffer, MAIN_Y + 10 + mode, MAIN_X - 6, 4, attr, 0, 0))
					{
					abort = 1;
					edit = 0;
					}
				*buf_start = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 2:	// Buffer End
				sprintf(buffer, "%04X", *buf_end);
				if(lined(buffer, MAIN_Y + 10 + mode, MAIN_X + 14, 4, attr, 0, 0))
					{
					abort = 1;
					edit = 0;
					}
				*buf_end = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 3:	// Fill
				if(mode == LOAD_BIN || mode == LOAD_HEX)
					{	
					sprintf(buffer, "%02X", *fill);
					if(lined(buffer, MAIN_Y + 10 + mode, MAIN_X + 35, 2, attr, 0, 1))
						{
						abort = 1;
						edit = 0;
						}
					*fill = strtol(buffer, NULL, 16);
					}
				++pos;
				break;
			case 4:				// Cancel
				setHighlite(MAIN_Y + 13 + mode, MAIN_X + 1, 8, attr, A_REVERSE);
				if(endMove())
					{
					edit = 0;
					status = 0;
					}
				else
					++pos;
				setHighlite(MAIN_Y + 13 + mode, MAIN_X + 1, 8, attr, A_NORMAL);
				break;
			case 5:				// Save/Load
				setHighlite(MAIN_Y + 13 + mode, MAIN_X + 22, 6, attr, A_REVERSE);
				if(endMove())
					{
					edit = 0;
					status = 1;
					}
				else
					pos = 0;
				setHighlite(MAIN_Y + 13 + mode, MAIN_X + 22, 6, attr, A_NORMAL);
				break;
			}
		}


	attroff(COLOR_PAIR(BLACK_WHITE));

	refresh();

	return(status);
	}


int	endMove()
	{
	int	loop = 1, status = 0;


	while(loop)
		{
		switch(getch())
			{
			case 9:			/* tab */
				loop = 0;
				status = 0;
				break;
			case 10:		/* enter */
				loop = 0;
				status = 1;
				break;
			default:
				break;
			}
		}
	
	return(status);
	}


int	lined(char *buffer, int y, int x, int max, int attr, int alpha_num, int disable)
	{
	int	i;
	int	ch, edit = 1, pos, len, update = 0, abort = 0, field_off = 0;
	char	alt_buf[1024];


	curs_set(2); // 0,1,2

	pos = len = strlen(buffer);

	attron(COLOR_PAIR(attr));

	while(edit)
		{
		if(update)
			{
			attron(COLOR_PAIR(attr));
			mvprintw(y, x, "%s", buffer);
			attroff(COLOR_PAIR(attr));
			update = 0;
			}
		move(y, x + pos);

		ch = getch();
		switch(ch)
			{
			case 27:		// ESC
				abort = 1;
			case 9:			/* tab */
			case 10:		/* enter */
				edit = 0;
				break;
			case 0x14:		// Cntl 't'
				if(disable)
					{
					if(field_off)
						{
						field_off = 0;
						strcpy(buffer,alt_buf);
						update = 1;
						}
					else
						{
						field_off = 1;
						strcpy(alt_buf, buffer);
						for(i = 0 ; i < max ; i++)
							*(buffer + i) = '-';

						*(buffer + i) = 0;
						update = 1;
						}
					}
				break;
			case 260:		/* left */
				if(field_off)
					break;

				if(pos)
					--pos;
				break;
			case 261:		/* right */
				if(field_off)
					break;

				if(pos < len)
					++pos;
				break;
			case 263:		/* backspace */
				if(field_off)
					break;

				if((len > 0) && (pos > 0))
					{
					backspace(buffer, pos, len);
					attron(COLOR_PAIR(attr));
					mvprintw(y, x + len -1, " ");
					attroff(COLOR_PAIR(attr));
					--len;
					--pos;
					update = 1;
					}
				break;
			default:
				if(field_off)
					break;

				if(alpha_num)
					{
					if((ch >= ' ') && (ch <= 'z') && len < max)
						{
						insert(ch, buffer, pos, len);
						++len;
						++pos;
						update = 1;
						}
					}
				else 
					{
					if(isxdigit(ch) && len < max)
						{
						ch = toupper(ch);
						insert(ch, buffer, pos, len);
						++len;
						++pos;
						update = 1;
						}
					}
				break;
			}
		}

	attroff(COLOR_PAIR(attr));

	curs_set(0); // 0,1,2

	if(field_off)
		strcpy(buffer, "-1");

	return(abort);
	}


int	insert(char ch, char *buffer, int pos, int len)
	{
	++len;

	while(len > pos)
		{
		*(buffer + len) = *(buffer + (len -1));
		--len;
		}

	*(buffer + pos) = ch;

	return(0);
	}


int	backspace(char *buffer, int pos, int len)
	{
	int	i;


	buffer += (pos-1);
	
	for(i = pos ; i <= len ; i++, buffer++)
		*buffer = *(buffer +1);


	return(0);		
	}


int	setHighlite(int y, int x, int len, int attr, int effect)
	{
	mvchgat(y, x, len, effect, attr, NULL);

	touchwin(stdscr);

	refresh();

	return(0);
	}


int	displayFile(int y, int x, int attr, char *filename)
	{
	int	len;
	char	tmp[32];


	len = strlen(filename);

	if(len > 13)
		filename += len - 13;

	sprintf(tmp, "[%-13s]", filename);

	attron(COLOR_PAIR(attr));

	mvprintw(y, x , tmp);

	attroff(COLOR_PAIR(attr));

	setHighlite(y, x + 1, 13, attr, A_BOLD);

	refresh();

	return(0);
	}


int	displayBufEnd(int y, int x, int attr, int buf_end)
	{
	char	tmp[16];


	sprintf(tmp, "[0x%04X]", buf_end);

	attron(COLOR_PAIR(attr));

	mvprintw(y, x , tmp);

	attroff(COLOR_PAIR(attr));

	setHighlite(y, x + 1, 6, attr, A_BOLD);

	refresh();

	return(0);
	}


int	menu_1(int cdypos, int cdxpos, int ylen, int xlen, int attr)
	{
        clear();

	mborder(cdypos, cdxpos, ylen, xlen, attr); 

	attron(COLOR_PAIR(attr));

	mvprintw(cdypos + 2,cdxpos + 6,"Load BIN");
	mvprintw(cdypos + 3,cdxpos + 6,"Load HEX");
	mvprintw(cdypos + 4,cdxpos + 6,"Save BIN");
	mvprintw(cdypos + 5,cdxpos + 6,"Save HEX");
	mvprintw(cdypos + 6,cdxpos + 6,"Read Elf MEMORY");
	mvprintw(cdypos + 7,cdxpos + 6,"Write Elf MEMORY");
	mvprintw(cdypos + 8,cdxpos + 6,"Clear Buffer");
	mvprintw(cdypos + 9,cdxpos + 6,"Edit Buffer");
	mvprintw(cdypos + 10,cdxpos + 6,"Fill Buffer");
	mvprintw(cdypos + 11,cdxpos + 6,"Run");
	mvprintw(cdypos + 12,cdxpos + 6,"Quit");

	attroff(COLOR_PAIR(attr));

	refresh();

	return(0);
	}


int	mborder(int cdypos, int cdxpos, int ylen, int xlen, int attr)
	{
	int	x1, x2, y1, y2;


	x1 = x2 = cdxpos;
	y1 = y2 = cdypos;

	x2 += xlen;
	y2 += ylen;

	attron(COLOR_PAIR(attr));

	while(cdypos <= y2)
		{
		move(cdypos ,cdxpos);

		while(cdxpos <= x2)
			{
			if(cdypos == y1)
				{
				if(cdxpos == x1)
					addch(ACS_ULCORNER);
				else if(cdxpos == x2)
					addch(ACS_URCORNER);
				else
					addch(ACS_HLINE);
				}
			else if(cdypos == y2)
				{
				if(cdxpos == x1)
					addch(ACS_LLCORNER);
				else if(cdxpos == x2)
					addch(ACS_LRCORNER);
				else
					addch(ACS_HLINE);
				}
			else
				{
				if(cdxpos == x1 || cdxpos == x2)
					addch(ACS_VLINE);
				else
					addch(' ');
				}

			++cdxpos;
			}

		++cdypos;
		cdxpos = x1;
		}

	attroff(COLOR_PAIR(attr));

	refresh();

	return(0);
	}


int	eborder(int cdypos, int cdxpos, int ylen, int xlen, int attr, int part1, int part2)
	{
	int	x1, x2, y1, y2;


	x1 = x2 = cdxpos;
	y1 = y2 = cdypos;

	x2 += xlen;
	y2 += ylen;

	attron(COLOR_PAIR(attr));

	while(cdypos <= y2)
		{
		move(cdypos ,cdxpos);

		while(cdxpos <= x2)
			{
			if(cdypos == y1)
				{
				if(cdxpos == x1)
					addch(ACS_ULCORNER);
				else if(cdxpos == x2)
					addch(ACS_URCORNER);
				else if(cdxpos == part1 || cdxpos == part2)
					addch(ACS_TTEE);
				else
					addch(ACS_HLINE);
				}
			else if(cdypos == y2)
				{
				if(cdxpos == x1)
					addch(ACS_LLCORNER);
				else if(cdxpos == x2)
					addch(ACS_LRCORNER);
				else if(cdxpos == part1 || cdxpos == part2)
					addch(ACS_BTEE);
				else
					addch(ACS_HLINE);
				}
			else
				{
				if(cdxpos == x1 || cdxpos == x2)
					addch(ACS_VLINE);
				else if(cdxpos == part1 || cdxpos == part2)
					addch(ACS_VLINE);
				else
					addch(' ');
				}

			++cdxpos;
			}

		++cdypos;
		cdxpos = x1;
		}

	attroff(COLOR_PAIR(attr));

	refresh();

	return(0);
	}


int	fillMbuffer(unsigned char *buffer, int start, int len, unsigned char fill)
	{
	int	max = MEMORY_SIZE;


	++max;
	++len;
	
	buffer += start;

	while(max-- && len--)
		*buffer++ = fill;

	return(len);
	}


int	fllBufferArea(int attr)
	{
	int	loop = 1, pos = 0;
	int	buf_start = 0, buf_end = 0, fill = 0;
	char	buffer[16];


	mborder(MAIN_Y + 2, MAIN_X - 13, 10, 54, attr);	// Edit Window

	mborder(MAIN_Y + 4, MAIN_X - 10, 2, 9, attr);	// Bstart
	mborder(MAIN_Y + 4, MAIN_X + 10, 2, 9, attr);	// Bend
	mborder(MAIN_Y + 4, MAIN_X + 30, 2, 9, attr);	// Fill

	mborder(MAIN_Y + 8, MAIN_X + 0, 2, 9, attr);	// Cancel
	mborder(MAIN_Y + 8, MAIN_X + 20, 2, 9, attr);	// Load/Save

	attron(COLOR_PAIR(attr));

	mvprintw(MAIN_Y + 3, MAIN_X - 10,"FILL START");
	mvprintw(MAIN_Y + 3, MAIN_X + 10,"FILL COUNT");
	mvprintw(MAIN_Y + 3, MAIN_X + 30,"FILL VALUE");

	mvprintw(MAIN_Y + 9, MAIN_X + 2,"CANCEL");

	mvprintw(MAIN_Y + 9, MAIN_X + 23,"FILL");


	// Fill in the Data

	sprintf(buffer, "0x%04X", buf_start);
	mvprintw(MAIN_Y + 5, MAIN_X - 8,buffer);

	sprintf(buffer, "0x%04X", buf_end);
	mvprintw(MAIN_Y + 5, MAIN_X + 12,buffer);

	sprintf(buffer, "0x%02X", fill);
	mvprintw(MAIN_Y + 5, MAIN_X + 33,buffer);

	refresh();

	while(loop)
		{
		switch(pos)
			{
			case 0:	// Buffer Start
				sprintf(buffer, "%04X", buf_start);
				if(lined(buffer, MAIN_Y + 5, MAIN_X - 6, 4, attr, 0, 0))
					loop = 0;
				buf_start = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 1:	// Buffer End
				sprintf(buffer, "%04X", buf_end);
				if(lined(buffer, MAIN_Y + 5, MAIN_X + 14, 4, attr, 0, 0))
					loop = 0;
				buf_end = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 2:	// Fill
				sprintf(buffer, "%02X", fill);
				if(lined(buffer, MAIN_Y + 5, MAIN_X + 35, 2, attr, 0, 0))
					loop = 0;
				fill = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 3:				// Cancel
				setHighlite(MAIN_Y + 9, MAIN_X + 1, 8, attr, A_REVERSE);
				if(endMove())
					loop = 0;
				else
					++pos;
				setHighlite(MAIN_Y + 9, MAIN_X + 1, 8, attr, A_NORMAL);
				break;
			case 4:				// Save/Load
				setHighlite(MAIN_Y + 9, MAIN_X + 22, 6, attr, A_REVERSE);
				if(endMove())
					{
					loop = 0;
					fillMbuffer(mbuffer, buf_start, buf_end, (unsigned char) fill);
					}
				else
					pos = 0;
				setHighlite(MAIN_Y + 9, MAIN_X + 22, 6, attr, A_NORMAL);
				break;
			}
		}

	attroff(COLOR_PAIR(BLACK_WHITE));

	refresh();

	return(0);
	}


int	clearBuffer(char *filename, int *buf_start, int *buf_end, int attr)
	{
	int	ch, loop = 1, pos = 0;


	mborder(6, 27, 10, 25, attr);
	mborder(12, 31, 2, 6, attr);
	mborder(12, 42, 2, 6, attr);

	attron(COLOR_PAIR(attr));

	mvprintw(9, 29,"CLEAR Buffer to 0x00 ?");
	mvprintw(13, 33,"NO");
	mvprintw(13, 44,"YES");

	setHighlite(13 + pos, 33, 3, attr, A_BOLD);

	refresh();

	getchar();


	while(loop)
		{
		ch = getch();

		switch(ch)
			{
			case 9:			/* tab */
				setHighlite(13, 33 + pos, 3, attr, A_NORMAL);

				if(pos)
					pos = 0;
				else
					pos = 11;

				setHighlite(13, 33 + pos, 3, attr, A_BOLD);
				break;
			case 10:		/* enter */
				loop = 0;

				if(pos == 11)
					{
					*filename = *buf_start = *buf_end = 0;
					fillMbuffer(mbuffer, 0, MEMORY_SIZE, 0x00);
					}
			case 27:		// ESC
				loop = 0;
				break;
			case 'n':
			case 'N':
				setHighlite(13, 33 + pos, 3, attr, A_NORMAL);

				pos = 0;

				setHighlite(13, 33 + pos, 3, attr, A_BOLD);
				break;
			case 'y':
			case 'Y':
				setHighlite(13, 33 + pos, 3, attr, A_NORMAL);

				pos = 11;

				setHighlite(13, 33 + pos, 3, attr, A_BOLD);
				break;
			case 260:		/* left */
				setHighlite(13, 33 + pos, 3, attr, A_NORMAL);

				if(pos)
					pos = 0;

				setHighlite(13, 33 + pos, 3, attr, A_BOLD);
				break;
			case 261:		/* right */
				setHighlite(13, 33 + pos, 3, attr, A_NORMAL);

				if(!pos)
					pos = 11;

				setHighlite(13, 33 + pos, 3, attr, A_BOLD);
				break;
			default:
				break;
			}

		}







/*
		switch(pos)
			{
			case 0:	// Filename
				if(lined(filename, MAIN_Y + 5 + mode, MAIN_X - 5, MAXFILENAME, attr, 1))
					{
					abort = 1;
					edit = 0;
					}
				++pos;
				break;
			case 1:	// Buffer Start
				sprintf(buffer, "%04X", *buf_start);
				if(lined(buffer, MAIN_Y + 10 + mode, MAIN_X - 6, 4, attr, 0))
					{
					abort = 1;
					edit = 0;
					}
				*buf_start = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 2:	// Buffer End
				sprintf(buffer, "%04X", *buf_end);
				if(lined(buffer, MAIN_Y + 10 + mode, MAIN_X + 14, 4, attr, 0))
					{
					abort = 1;
					edit = 0;
					}
				*buf_end = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 3:	// Fill
				sprintf(buffer, "%02X", *fill);
				if(lined(buffer, MAIN_Y + 10 + mode, MAIN_X + 35, 2, attr, 0))
					{
					abort = 1;
					edit = 0;
					}
				*fill = strtol(buffer, NULL, 16);
				++pos;
				break;
			case 4:				// Cancel
				setHighlite(MAIN_Y + 13 + mode, MAIN_X + 1, 8, attr, A_REVERSE);
				if(endMove())
					{
					edit = 0;
					status = 0;
					}
				else
					++pos;
				setHighlite(MAIN_Y + 13 + mode, MAIN_X + 1, 8, attr, A_NORMAL);
				break;
			case 5:				// Save/Load
				if(endMove())
					{
					edit = 0;
					status = 1;
					}
				else
					pos = 0;
				setHighlite(MAIN_Y + 13 + mode, MAIN_X + 22, 6, attr, A_NORMAL);
				break;
			}
		}
*/


	attroff(COLOR_PAIR(attr));

	refresh();

	return(0);
	}



/* ************************************************************************* */
/* ***************************** HEX EDITOR Functions ********************** */
/* ************************************************************************* */


int	hexed(unsigned char *image, int *length)
	{
	int	run = 1, update = 1, edit = 0, jadd;
	int	y = 0, x = 0, nibble = 0;
	int	offset = 0;
	int	start = -1, end = -1;


        clear();
	eborder(3, 2, 17, 76, WHITE_BLACK, 9, 59); 
	curs_set(2); // 0,1,2

	// do border

	while(run)
		{
		if(update)
			{
			update = 0;
			printimage(image, offset);
			}

		mvprintw(21, 4, "%s  Current", hexstr(offset + x + (y * 16), 4));

		if(!edit)
			attron(A_DIM);

		mvprintw(22, 4, "%s  End", hexstr(*length, 4));

		if(!edit)
			attroff(A_DIM);

		move(4 + y, 11 + (3 * x) + nibble);

		refresh();

		switch(getch())
			{
			case 47:		/* SLASH */
				/* Locate */
				break;
			case 'g':
				mvprintw(21, 20, "GoTo -> ");
				jadd = getaddress(4, 21, 28);
				mvprintw(21, 20, "                ");

				if(jadd >= MEMORY_SIZE - 256)
					jadd = 0xFF00;
				if(jadd != -1)
					{
					offset = jadd & (~0xF);
					x = jadd & 0xF;
					y = nibble = 0;
					update = 1;
					}
				break;
			case 'h':		/* Help Screen */
				help();
				eborder(3, 2, 17, 76, WHITE_BLACK, 9, 59); 
				update = 1;
				break;
			case 'p':		/* Place */
				/* Fill - what and how long */
				break;
			case 'q':		/* Quit */
			case 'Q':		/* Quit */
				run = 0;
				break;
			case 's':		/* Set Mark */
				start = offset+x+(y*16);
				break;
			case 'S':		/* End Set */
				end = offset+x+(y*16);
				break;
			case 19:		/* Clear marks */
				start = end = -1;
				break;
			case '0':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x0 : 0x00);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '1':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x1 : 0x10);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '2':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x2 : 0x20);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '3':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x3 : 0x30);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '4':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x4 : 0x40);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '5':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x5 : 0x50);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '6':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x6 : 0x60);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '7':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x7 : 0x70);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '8':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x8 : 0x80);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case '9':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0x9 : 0x90);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case 'a':
			case 'A':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0xA : 0xA0);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case 'b':
			case 'B':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0xB : 0xB0);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case 'c':
			case 'C':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0xC : 0xC0);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case 'd':
			case 'D':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0xD : 0xD0);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case 'e':
			case 'E':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0xE : 0xE0);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case 'f':
			case 'F':
				*(image+offset+x+(y*16)) = (*(image+offset+x+(y*16)) & (nibble ? 0xF0 : 0xF)) | (nibble ? 0xF : 0xF0);
				if(*length < offset+x+(y*16))
					*length = offset+x+(y*16);
				cursor_right(&nibble, &y, &x, &offset);
				update = 1;
				edit = 1;
				break;
			case 258:		/* DOWN */
				if(y < 15)
					++y;
				else if(offset < (MEMORY_SIZE - 256))
					{
					offset += 16;
					update = 1;
					}
				break;
			case 259:		/* UP */
				if(y)
					--y;
				else if(offset)
					{
					offset -= 16;
					update = 1;
					}
				break;
			case 261:		/* RIGHT */
				if(nibble)
					{
					nibble = 0;
					if(x < 15)
						++x;
					else if(y < 15)
						{
						++y;
						x = 0;
						}
					else if(offset < (MEMORY_SIZE - 256))
						{
						x = 0;
						update = 1;
						offset += 16;
						}
					}
				else
					nibble = 1;

				break;
			case 260:		/* LEFT */
				if(nibble)
					nibble = 0;
				else if(x)
					{
					--x;
					nibble = 1;
					}
				else if(y)
					{
					--y;
					x = 15;
					nibble = 1;
					}
				else if(offset)
					{
					x = 15;
					update = 1;
					nibble = 1;
					offset -= 16;
					}
				break;
			case 262:		/* HOME */
				offset = y = x = nibble = 0;
				update = 1;
				break;
			case 338:		/* PgDOWN */
				if((offset + 256) < (MEMORY_SIZE - 256))
					{
					offset += 256;
					update = 1;
					}
				break;
			case 339:		/* PgUP */
				if((offset - 256) >= 0)
					offset -= 256;
				else
					offset = 0;
				update = 1;
				break;
			case 360:		/* END */
				offset = *length & (~0xF);
				x = *length & 0xF;
				y = nibble = 0;
				update = 1;
				break;
			default:
				break;
			}
		}

	curs_set(0); // 0,1,2

	clear();

	refresh();

	return(0);
	}


int	getaddress(int max, int y, int x)
	{
	int	ch, count = 0, loop = 1;
	char	buffer[32];
	static	char	clrstr[17] = {"                "};


	buffer[0] = 0;

	while(loop)
		{
		mvprintw(y, x, "%s", clrstr);
		mvprintw(y, x, "%s", buffer);
		ch = getch();
		switch(ch)
			{
			case 27:
				return(-1);
			case 10:
				if(buffer[0] == 0)
					return(-1);
				else
					return(strtol(buffer, (char **)NULL, 16));
			case 263:
				if(count)
					buffer[--count] = 0;
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case 'a':
			case 'A':
			case 'b':
			case 'B':
			case 'c':
			case 'C':
			case 'd':
			case 'D':
			case 'e':
			case 'E':
			case 'f':
			case 'F':
				if(count < max)
					{
					buffer[count++] = ch;
					buffer[count] = 0;
					}
				break;
			default:
				break;
			}
		}

	return(0);	
	}


int	cursor_right(int *nibble, int *y, int *x, int *offset)
	{
	if(*nibble)
		{
		*nibble = 0;
		if(*x < 15)
			++(*x);
		else if(*y < 15)
			{
			++(*y);
			*x = 0;
			}
		else if(*offset < (MEMORY_SIZE - 256))
			{
			*x = 0;
			*offset += 16;
			}
		}
	else
		*nibble = 1;

	return(0);
	}

	
int	printimage(unsigned char *image, int offset)
	{
	int	x, y;


	for(y = 0 ; y < 16 ; y++)
		{
		mvprintw(y + 4, 4, "%s", hexstr(offset, 4));
		for(x = 0 ; x < 16 ; x++)
			{
			mvprintw(y + 4, 61 + x, "%c", (isprint(*(image + offset)) ? *(image + offset) : '.'));
			mvprintw(y + 4, 11 + (x * 3), "%s", hexstr(*(image + offset++),2));
			}
		}

	refresh();

	return(0);
	}


char	*hexstr(int value, int digits)
	{
	int		i = 7;
	static	char	str[9];
	static	char	chars[17] = {"0123456789ABCDEF"};


	str[0] = str[1] = str[2] = str[3] = str[4] = str[5] = str[6] = str[7] = '0';
	str[8] = 0;

	while(value)
		{
		str[i--] = chars[value & 0xf];
		value = value >> 4;
		}

	return(str + (8 - digits));
	}


void	help()
	{
        clear();
	mborder(5, 20, 9, 40, 1); 
	move(26,40);
        refresh();

	attron(COLOR_PAIR(1));

	mvprintw(7,  22, "/<string>          Locate <string>");
	mvprintw(8,  22, "g<address>         GoTo <address>");
	mvprintw(9,  22, "h                  Help page");
	mvprintw(10, 22, "p <ch> <count>     Place <ch> <count>");
	mvprintw(11, 22, "q/Q                Quit");
	mvprintw(12, 22, "Cntl s             Set Buffer End");

	attroff(COLOR_PAIR(1));

	getch();

        clear();
        refresh();
	}


